import torch
import torch.nn as nn
import os
import dgl
import numpy as np


def cuda(x):
    if torch.cuda.is_available() and not os.getenv('NOCUDA', None):
        return x.to(torch.device('cuda'))   # works for both DGLGraph and tensor
    else:
        return x


class GRUUpdate(nn.Module):
    def __init__(self, hidden_size):
        nn.Module.__init__(self)
        self.hidden_size = hidden_size

        self.W_z = nn.Linear(2 * hidden_size, hidden_size)
        self.W_r = nn.Linear(hidden_size, hidden_size, bias=False)
        self.U_r = nn.Linear(hidden_size, hidden_size)
        self.W_h = nn.Linear(2 * hidden_size, hidden_size)

    def update_zm(self, node):
        src_x = node.data['src_x']
        s = node.data['s']
        rm = node.data['accum_rm']
        z = torch.sigmoid(self.W_z(torch.cat([src_x, s], 1)))
        m = torch.tanh(self.W_h(torch.cat([src_x, rm], 1)))
        m = (1 - z) * s + z * m
        # print('input: ', torch.cat([src_x, s], 1))
        # print('self.W_z: ', self.W_z.weight)
        if torch.isnan(s).any() or torch.isnan(rm).any() or torch.isnan(z).any() or torch.isnan(m).any():
            print('in update_zm:')
            breakpoint()
        return {'m': m, 'z': z}

    def update_r(self, node, zm=None):
        dst_x = node.data['dst_x']
        m = node.data['m'] if zm is None else zm['m']
        r_1 = self.W_r(dst_x)
        r_2 = self.U_r(m)
        r = torch.sigmoid(r_1 + r_2)
        if torch.isnan(m).any() or torch.isnan(r_1).any() or torch.isnan(r_2).any() or torch.isnan(r).any():
            print('in update_r:')
            breakpoint()
        return {'r': r, 'rm': r * m}

    def forward(self, node):
        dic = self.update_zm(node)
        dic.update(self.update_r(node, zm=dic))
        return dic

def tocpu(g):
    src, dst = g.edges()
    src = src.cpu()
    dst = dst.cpu()
    return dgl.graph((src, dst), num_nodes=g.number_of_nodes())

def graphs_batch_offsets(graphs_batch):
    if torch.cuda.is_available() and not os.getenv('NOCUDA', None):
        return np.cumsum(np.insert(graphs_batch.batch_num_nodes().cpu().numpy(), 0, 0))
    else:
        return np.cumsum(np.insert(np.array(graphs_batch.batch_num_nodes), 0, 0))


class BinaryClassificationMeter(object):
    """Computes and stores the average and current value"""
    def __init__(self):
        self.reset()

    def reset(self):
        self.tp = 0
        self.tn = 0
        self.fp = 0
        self.fn = 0
        self.acc = 0
        self.pre = 0
        self.rec = 0
        self.f1 = 0

    def update(self, output, target):
        pred = output >= 0.5
        pred = pred.type(torch.IntTensor)

        truth = target >= 0.5
        truth = truth.type(torch.IntTensor)

        self.tp += pred.mul(truth).sum(0).float()
        self.tn += (1 - pred).mul(1 - truth).sum(0).float()
        self.fp += pred.mul(1 - truth).sum(0).float()
        self.fn += (1 - pred).mul(truth).sum(0).float()
        self.acc = (self.tp + self.tn).sum() / (self.tp + self.tn + self.fp + self.fn).sum()
        self.pre = self.tp / (self.tp + self.fp)
        self.rec = self.tp / (self.tp + self.fn)
        self.f1 = (2.0 * self.tp) / (2.0 * self.tp + self.fp + self.fn)
        self.avg_pre = np.nanmean(self.pre)
        self.avg_rec = np.nanmean(self.rec)
        self.avg_f1 = np.nanmean(self.f1)